#include <cmath>
#include <cstdlib>
#include <iostream>
#include <vector>

using namespace std;

#include "points3.h"
#include "simulate.h"
#include "perlin-noise.h"

namespace Hex {
  float neighbor_dx[7], neighbor_dy[7];
}

/* Notes about water

Should water flow equations take into account narrow canyons forcing
water to be up higher?

River edges look very straight.  We might want to multitexture using
an alpha layer with non-straight edges.  That way rivers would look a
little more varied and natural.

We should store the min/max z of each hexagon somewhere.  This could
be used for water rendering (comments are in the water-style 4 section
of the code.)

Should water flow into a hexagon based on its min z value instead of
center z value?

Once you grab onto a hex and start raising/lowering, movement of the
mouse should move to adjacent hexes instead of using picking

Moisture and vegetation have interesting rules:
  1. Water kills vegetation
  2. Water maxes out moisture
  3. Moisture spreads to adjacent hexes
  4. Vegetation grows in response to moisture
So you end up with a system where vegetation grows near but not in water.

TODO: moisture shouldn't spread to adjacent hexes that are higher up

TODO: smoothing of land should depend on the vegetation level

TODO: smoothing of land should depend on the moisture/water level
(moist land erodes quicker; watery land erodes much quicker)

TODO: erosion should depend on the vegetation/moisture level

*/

/* Simulation phases should be objects that are parameterized on
   neighbor validity tests.  That way we can factor out the &1 for
   neighbor coordinates and we can also factor out the Map::valid
   tests. */

namespace Rendering {
  // TODO: replace this with 'game time'
  extern int frame;
}

namespace Map {
  bool flooding = false;
  float erosion_rate = 0.01;
  float total_vegetation = 0.0f;

  // The main height map comes with two derived vector maps: normals
  // (perpendicular to the land) and gradients (downwards-pointing)
  array height;
  array nx_base, ny_base, nz_base; // vertex normals
  array gx_base, gy_base, gz_base; // vertex gradients

  array moisture;
  array vegetation;
  array water;
  array outflow;
  array silt;
  
  Vector3d water_momentum[size][size];
  vector<WaterSource> water_sources;
  
  void water_flow();
  
  void water_flow_from_sources() {
    float scale = flooding? 1 : 0.02;
    for (vector<Map::WaterSource>::iterator i = Map::water_sources.begin();
         i != Map::water_sources.end(); ++i) {
      Map::water[i->m][i->n] += scale * (1.5 + cos(Rendering::frame*i->frequency));
    }
  }
  
  void smooth_value(array values, float alpha, float threshold) {
    LOOP(m, n) {
      float x = values[m][n];
      float total = 0.0f;
      float count = 0.0f;
      bool smoothing = false;
      for (int k = 1; k <= 6; ++k) {
        int m1 = m + Hex::dm[k];
        int n1 = n + Hex::dn[m&1][k];
        if (Map::valid(m1, n1)) {
          float x1 = values[m1][n1];
          total += x1;
          count += 1.0f;
          if (x1-x > threshold || x-x1 > threshold) {
            smoothing = true;
          }
        }
      }

      if (smoothing) {
        values[m][n] = alpha*(total/count) + (1.0f-alpha)*x;
      }
    }
  }
  
  void setup_map() {
    LOOP(m, n) {
      height[m][n] = 0.0;
      water[m][n] = 0.0;
      silt[m][n] = 0.0;
      outflow[m][n] = 0.0;
      moisture[m][n] = 0.0;
      vegetation[m][n] = 0.01;
    }

    // Two overlapping maps that are not aligned .. to avoid some artifacts
    PerlinNoise p, q, r;
    p.reseed(); q.reseed(); r.reseed();

#if 1
    // This is pretty wacky.. I'm not sure what I'm doing.
    for (int m = 0; m < size; ++m) {
      for (int n = 0; n < size; ++n) {
        float altitude = 15.0f * q.noise(5.0f*m/size, 5.0f*n/size);
        height[m][n] += altitude;

        float x = 3.0f*(float(m)/size-0.5f), y = 3.0f*(float(n)/size-0.5f);
        float rotation_x = r.noise(0.2f + x, 0.3f + y);
        float rotation_y = sqrt(1.0f - rotation_x*rotation_x);
        float x2 = x*rotation_x + y*rotation_y;
        float y2 = y*rotation_x + x*rotation_y;
        float scale = 0.8f * (1.0f + 5.0f*r.noise(x2, y2));
        height[m][n] += max(0.2f, 3.5f+height[m][n])*2.8f * (p.noise(scale*m, scale*n) - 0.2f);

      }
    }
#endif
    
    for (int hill = 0; hill < 10; ++hill) {
      int m0 = random() % size, n0 = random() % size;

      for (int m = m0 - 10; m <= m0 + 10; ++m) {
        for (int n = n0 - 10; n <= n0 + 10; ++n) {
          if (valid(m, n)) {
            height[m][n] += 125.0 / (10 + m*m + n*n);
          }
        }
      }
    }
    
    for (int m = 0; m < size; ++m) {
      for (int n = 0; n < size; ++n) {
        
        /* Raise the very edges of the map even more */
        if (m <= 2 || m >= size-3 || n <= 2 || n >= size-3)
          height[m][n] = 15;
      }
    }

#if 0
    int road_direction = 1 + random()%6;
    int road_m = size/2 + random()%20 - random()%20;
    int road_n = size/2 + random()%20 - random()%20;
    water_sources.push_back(WaterSource(road_m, road_n, 0.05));
    float road_altitude = height[road_m][road_n];
    int run_length = random()%60;
    for (int i = 0; i < 300; ++i) {
      /* Move to another spot */
      road_n += Hex::dn[road_m&1][road_direction];
      road_m += Hex::dm[road_direction];
      /* Possibly change direction */
      do {
        if (--run_length <= 0) {
          road_direction += 1 - 2*(random()%2);
          run_length = 5 + random() % 60;
        }
        road_direction = 1 + (road_direction + 5) % 6;
      } while (!valid(road_m + Hex::dm[road_direction],
                           road_n + Hex::dn[road_m&1][road_direction]));
      /* Change altitude */
      float const alpha = 0.05;
      road_altitude = (1-alpha)*road_altitude + alpha*height[road_m][road_n];
      for (int k = 0; k < 7; ++k) {
        /* Don't alter the land in the road_direction because we need to know
           its original height for the road_altitude calculation. */
        if (k == road_direction) continue;
        int m = road_m + Hex::dm[k];
        int n = road_n + Hex::dn[road_m&1][k];
        if (valid(m, n)) {
          height[m][n] = road_altitude;
        }
      }
    }
#endif
    
    water_sources.push_back(WaterSource(70, 3, 0.0013));
    water_sources.push_back(WaterSource(3, 50, 0.001));
    water_sources.push_back(WaterSource(3, 120, 0.01));

    for (vector<WaterSource>::iterator i = water_sources.begin();
         i != water_sources.end(); ++i) {
      // Change the location of the springs to climb up hills.  We use
      // something like simulated annealing, where we add randomness
      // to allow the spring to move to less appealing spots at first
      // but later on it's not allowed to move so freely.
      for (int j = 0; j < 40; ++j) {
        for (int k = 1; k <= 6; ++k) {
          int m1 = i->m + Hex::dm[k];
          int n1 = i->n + Hex::dn[i->m&1][k];
          if (valid(m1, n1) && m1 > 3 && n1 > 3 &&
              m1 < size-3 && n1 < size-3) {
            if (height[m1][n1] + (random()%100)/(2.0f+j)
                > height[i->m][i->n]) {
              i->m = m1;
              i->n = n1;
              break;
            }
          }
        }
      }
    }

    LOOP(m,n) {
      water[m][n] = 1.0;
    }
    erosion_rate = 3.0;
    flooding = true;
    float avg_height = 0.0;
    float max_height = -999.0;
    float min_height = +999.0;
    for (int i = 150; i >= 0; --i) {
      float total_height = 0.0;
      LOOP(m,n) {
        total_height += height[m][n];
        max_height = max(max_height, height[m][n]);
        min_height = min(min_height, height[m][n]);
        height[m][n] -= avg_height;
        height[m][n] *= 1.004;
        height[m][n] += 0.01*water[m][n] + 0.0001*m;
        // Uplift
        height[m][n] += 0.05*(size-m + size-n)/size;
      }
      avg_height = total_height / size / size;
      max_height = 0.95 * max_height;
      min_height = 0.95 * min_height;
      water_flow_from_sources();
      water_flow();
      if (i > 50) {
        smooth_value(height, 0.01f, 1.0f); // lots of smoothing early on
      }
      erosion_rate = 1.001*erosion_rate;

      if (i % 50 == 0) {
        cout << "  Erosion - " << i << " iterations left." << endl;
        cout << "     heights: "
             << " min = " << min_height
             << " avg = " << avg_height
             << " max = " << max_height
             << endl;
        float const rescale = (i > 50)? 10.0f : 15.0f;
        LOOP(m,n) {
          if (height[m][n] > avg_height) {
            height[m][n] = +rescale/(max_height-avg_height)*(height[m][n]-avg_height);
          } else {
            height[m][n] = -rescale/(min_height-avg_height)*(height[m][n]-avg_height);
          }
          
          water[m][n] = 0.0;
          silt[m][n] = 0.0;
        }
        for (int i = 0; i < 50; ++i) {
          water_flow_from_sources();
          water_flow();
        }
      }
    }
    flooding = false;
    erosion_rate = 0.01;
    
#if 1
    // Plateaus around each water source
    LOOP(m, n) {
      for (vector<Map::WaterSource>::iterator i = Map::water_sources.begin();
           i != Map::water_sources.end(); ++i) {
        int dm = m - i->m;
        int dn = n - i->n;
        if (dm*dm + dn*dn < 100) {
          height[m][n] = 7 + height[m][n]/2;
        }
      }
    }
#endif
    
  }

  void update_vegetation_and_silt(int m, int n) {
    // TODO: it's bad to base anything on length here because it's
    // length in PIXELS!
    // total_vegetation += Map::vegetation[m][n];

    if (Map::water[m][n] > 0) {
      float speed_squared = Map::water_momentum[m][n].length_squared();
    
      Map::water_momentum[m][n] *= 0.8f;
      
      // TODO: what about depositing into neighboring hexes?
      // TODO: what about not taking silt if neighbors are a lot higher?
      // TODO: what about meandering?
      float max_silt = min(1.0f,
                           50.0f*erosion_rate
                           *(1.1f-Map::moisture[m][n])
                           *(1.1f-Map::vegetation[m][n])
                           *(0.001f + speed_squared));
      float dsilt = 0.01f * (max_silt - Map::silt[m][n]);
      Map::silt[m][n] += dsilt;
      Map::height[m][n] -= dsilt;
      // TODO: maybe shallow water doesn't kill vegetation as quickly
      Map::vegetation[m][n] *= 0.99f;
      Map::moisture[m][n] += 0.1f*Map::water[m][n];
      if (Map::moisture[m][n] > 1.0f) Map::moisture[m][n] = 1.0f;
    } else {
      // Drop ALL silt right away
      Map::height[m][n] += Map::silt[m][n];
      Map::silt[m][n] = 0.0f;
      // Regrowth:
      Map::vegetation[m][n] = 0.999f*Map::vegetation[m][n]
                              + 0.002f*Map::moisture[m][n];
      // Drying:
      Map::moisture[m][n] *= 0.999f;
      // No water means no momentum
      Map::water_momentum[m][n].x = 0.0f;
      Map::water_momentum[m][n].y = 0.0f;
      Map::water_momentum[m][n].z = 0.0f;
    }
  }
  
  /* TODO: Water flow is a pretty expensive part of the simulation.
     It only runs for parts of the world that have water.  Most of the
     time, that's small, so it's no problem, but occasionally it'll be
     large.  So we need to rate limit this by having it compute a
     certain number of hexes per tick, and it should be able to resume
     at some row. */

  // Water simulation based on 10 July 2005 notes.  There are two
  // components: depth (Map::water) and velocity (Map::vx, vy, vz).
  // At every step we compute forces at every point.  The force is
  // based on gravity (difference in Map::height) and water pressure
  // (difference in Map::water).  We actually use acceleration instead
  // of force, because it works independent of mass.  Once we have
  // velocity vectors, we compute the new location of each mass of
  // water, and move it.  The new location will be a point within a
  // triangle; we compute the barycentric coordinates of that triangle
  // and move water to the three vertices.  TODO: this algorithm is
  // unstable and unusable.

array vx, vy; // Velocity vector for water, in world coordinates

  // TODO: merge water_momentum and vx, vy, vz

// If the coordinates are out of bounds, reflect them across the wall
// to another point location is in bounds
void reflect(int& w, int w_min, int w_max) {
  if (w < w_min) {
    w = 2*w_min - w;
  }
  if (w > w_max) {
    w = 2*w_max - w;
  }
  if (w < w_min) {
    w = w_min; // just in case w is really far away from w_min..w_max
  }
}

void reflect_coordinates(int& m, int& n) {
  reflect(m, 1, Map::size-2);
  reflect(n, 1, Map::size-2);
}

  void water_flow_new() {
    const float dt = 0.1f;

    // First pass: modify the velocity vectors by the acceleration
    
    LOOP(m, n) {
      if (Map::water[m][n] == 0.0f) continue;
      
      // TODO: we only need k <= 3, because the forces are symmetrical, right?
      for (int k = 1; k <= 6; ++k) {
        int dm = Hex::dm[k], dn = Hex::dn[m&1][k];
        if (!Map::valid(m+dm, n+dn)) continue;
      
        float dw = Map::water[m+dm][n+dn] - Map::water[m][n];
        float dh = Map::height[m+dm][n+dn] - Map::height[m][n];

        float acceleration = dw + dh;
        vx[m][n] += dt * acceleration * Hex::neighbor_dx[k];
        vy[m][n] += dt * acceleration * Hex::neighbor_dy[k];
      }
    }

    // Second pass: move water along the velocity vectors

    array nw; // New water level, used temporarily
    array nvx, nvy; // New velocities, used temporarily
    LOOP(m, n) {
      nw[m][n] = 0.0f;
      nvx[m][n] = 0.0f;
      nvy[m][n] = 0.0f;
    }
    LOOP(m, n) {
      if (Map::water[m][n] == 0.0f) continue;

      Point3d p = grid_to_world(m, n); // TODO: use Rendering::x,y
      barycentric_location B(p.x + vx[m][n], p.y + vy[m][n]);

      // If our new coordinates aren't in range, pick a point in range
      reflect_coordinates(B.m1, B.n1);
      reflect_coordinates(B.m2, B.n2);
      reflect_coordinates(B.m3, B.n3);

      // Transfer everything from m,n to the new coordinates, weighted
      // by the barycentric coordinates
      nw[B.m1][B.n1] += B.p1 * Map::water[m][n];
      nw[B.m2][B.n2] += B.p2 * Map::water[m][n];
      nw[B.m3][B.n3] += B.p3 * Map::water[m][n];
      nvx[B.m1][B.n1] += B.p1 * vx[m][n];
      nvx[B.m2][B.n2] += B.p2 * vx[m][n];
      nvx[B.m3][B.n3] += B.p3 * vx[m][n];
      nvy[B.m1][B.n1] += B.p1 * vy[m][n];
      nvy[B.m2][B.n2] += B.p2 * vy[m][n];
      nvy[B.m3][B.n3] += B.p3 * vy[m][n];
    }
    LOOP(m, n) {
      const float alpha = 1.0f;
      // TODO: it would be nicer if we could swap arrays here
      Map::water[m][n] = nw[m][n] * alpha + Map::water[m][n] * (1.0f - alpha);
      vx[m][n] = nvx[m][n] * alpha + vx[m][n] * (1.0f - alpha);
      vy[m][n] = nvy[m][n] * alpha + vy[m][n] * (1.0f - alpha);
    }

    // Compatibility. TODO: remove me
    LOOP(m, n) {
      Map::water_momentum[m][n] = Vector3d(vx[m][n], vy[m][n], 0.0);
      vx[m][n] *= 0.99f;
      vy[m][n] *= 0.99f;
    }

    smooth_value(Map::water, 0.01f, 3.0f);
    smooth_value(vx, 0.01f, 0.0f);
    smooth_value(vy, 0.01f, 0.0f);
  }
  
  void water_flow() {
    float const OUTFLOW_ALPHA = 0.2f;

    if (Rendering::frame % 4 == 0) {
      LOOP(m,n) {
        update_vegetation_and_silt(m, n);
      }
    }
    
    LOOP(m, n) {
      if (Map::water[m][n] < 0.001f) {
        Map::water[m][n] = 0.0f;
        continue;
      }
      
      int best_k = -1;
      float best_w = 0.0f;
      float best_h = 0.0f;
      float best_rank = 0.0f;

      for (int k = 1; k <= 6; ++k) {
        int dm = Hex::dm[k], dn = Hex::dn[m&1][k];
        if (!Map::valid(m+dm, n+dn)) continue;
      
        float dh = Map::water[m][n]+Map::height[m][n]
                   - (Map::water[m+dm][n+dn]+Map::height[m+dm][n+dn]
                      - Map::outflow[m+dm][n+dn]);
        float bonus = 0;

        if (Map::water[m+dm][n+dn] > 0.001f)
          bonus += 0.01f;

        Vector3d angle(Hex::neighbor_dx[k], Hex::neighbor_dy[k], 0.0f);
        bonus += max(0.0f, 0.005f*dot(angle, angle+0.1f*Map::water[m][n]*Map::water_momentum[m][n]));
        bonus += max(0.0f, 0.005f*dot(angle, angle+0.05f*Map::water[m+dm][n+dn]*Map::water_momentum[m+dm][n+dn]));

        
        bonus += 0.01f * (1.0f - Map::vegetation[m+dm][n+dn]); // easier to flow into non-vegetation. TODO: use moisture too?

        if (dh > 0.0f) {
          // Flow from one to another NOTE: these two constants are the
          // result of much experimentation .. but no good conclusion
          // reached
          float w = min(Map::water[m][n]*0.3f, dh*0.15f);
          float rank = w + bonus;

          if (rank > best_rank) {
            best_k = k;
            best_w = w;
            best_h = dh;
            best_rank = rank;
          }
        }
      }

      if (best_k != -1) {
        int dm = Hex::dm[best_k], dn = Hex::dn[m&1][best_k];
        float dsilt = 0.0f;
        // Carry over some silt from the previous spot
        dsilt += Map::silt[m][n];
        // Scale down by the fraction of water we're transferring
        dsilt *= best_w / Map::water[m][n];
        Map::silt[m][n] -= dsilt;
        Map::silt[m+dm][n+dn] += dsilt;
        Map::water[m][n] -= best_w;
        Map::water[m+dm][n+dn] += best_w;
        Map::outflow[m][n] = (1.0f-OUTFLOW_ALPHA) * Map::outflow[m][n]
                             + OUTFLOW_ALPHA * best_w;

        // Momentum is the product of mass (water) and velocity.  It is
        // being transferred from the first hex to the second.
        Vector3d angle(Hex::neighbor_dx[best_k], Hex::neighbor_dy[best_k], 0.0f);
        Map::water_momentum[m][n] *= Map::water[m][n] / (best_w + Map::water[m][n]);
        Map::water_momentum[m+dm][n+dn] += best_h * best_w / (Map::water[m+dm][n+dn]) * angle;
      }

      if (best_k == -1) {
        // No water flow, but there was water.
        Map::outflow[m][n] *= (1.0f-OUTFLOW_ALPHA);
      }
    
      /* Evaporation/absorption - only for stationary water */
      if (best_w < 0.03f) {
        // Vegetation helps absorb and evaporate water.  Or should dry
        // climates increase evaporation?  I don't know.
        float const evaporation_speed = 0.0001f * (1.0f + 3.0f * Map::vegetation[m][n]);
        if (Map::water[m][n] >= evaporation_speed) {
          Map::water[m][n] -= evaporation_speed;
        } else if (Map::water[m][n] > 0.0f) {
          Map::water[m][n] = 0.0f;
        }
      }
    }
  }

  void simulate() {
    water_flow_from_sources();
    water_flow();
    if (Rendering::frame % 4 == 0) {
      smooth_value(Map::moisture, 0.06f, 0.005f);
      smooth_value(Map::vegetation, 0.1f, 0.0f);
    }
    if (Rendering::frame % 32 == 2) {
      smooth_value(Map::height, 0.01f, 5.0f);
    }
  }
  
  
}
