#include "include_gl.h"
#include <iostream>

#include "textures.h"
using namespace std;

#include "perlin-noise.h"

namespace Texture {
  GLuint polished;
  GLuint smooth;
  GLuint rough;
  GLuint contour;
  GLuint road;

  void setup_generic(GLuint& texture, float texture_scale, float variance) {
    int const W = 256, H = 256;
    unsigned char bytes[W*H];
    float const T = texture_scale;

    // This texture is generic so that we can apply it with many
    // different colors.
    PerlinNoise p;
    p.reseed();
    
    for (int i = 0; i < W; ++i) {
      for (int j = 0; j < H; ++j) {
        float x = T * i / W;
        float y = T * j / H;

        // This makes Perlin noise tilable
        float v = p.noise(x, y) * (T-x) * (T-y) +
                  p.noise(x-T, y) * x * (T-y) +
                  p.noise(x-T, y-T) * x * y +
                  p.noise(x, y-T) * (T-x) * y;
        v = v / (T*T);
        bytes[i*W+j] = int(255*(1.0f-variance + variance*0.5f*(v+1.0f)));
      }
    }
    // Ideas from Ken Perlin: use multiple noise functions at
    // different scales.  His "turbulence" takes the abs(noisefn)
    // at multiple scales, where the abs() creates some
    // discontinuities.  Or use the noise function for phase
    // shifting something else.  Use a 3d noise function to
    // generate multiple textures over time (3rd dimension is
    // time).  Flames: use noise in 2d to phase shift a yellow to
    // red gradient, and use noise in 3rd dimension for animation.
    // Create cycles by F2(x,y,z) = (1-z)*F(x,y,z) + z*F(x,y,z-1).
    // See http://www.noisemachine.com/talk1/ for the talk.
    // http://students.vassar.edu/mazucker/code/perlin-noise-math-faq.html
        
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_REPEAT);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_LUMINANCE,
                 W, H, 0,
                 GL_LUMINANCE, GL_UNSIGNED_BYTE, bytes);
  }

  void setup_volume(GLuint& texture) {
    int const W = 32;
    unsigned char bytes[W*W*W];
    float const T = 4.0f;

    PerlinNoise p;
    p.reseed();
    
    for (int i = 0; i < W; ++i) {
      for (int j = 0; j < W; ++j) {
        for (int k = 0; k < W; ++k) { // depth
          float x = T * i / W;
          float y = T * j / W;
          float z = T * k / W;
          
          // This makes Perlin noise tilable
          float v = 0.0;
          for (int flip_x = 0; flip_x < 2; ++flip_x) {
            for (int flip_y = 0; flip_y < 2; ++flip_y) {
              for (int flip_z = 0; flip_z < 2; ++flip_z) {
                v += p.noise( flip_x?x:x-T, flip_y?y:y-T, flip_z?z:z-T )
                  * (flip_x?(T-x):x)
                  * (flip_y?(T-y):y)
                  * (flip_z?(T-z):z);
              }
            }
          }
          v = v / (T*T*T);
          bytes[i*W*W+j*W+k] = max(0, min(255, int(200 + v*55.0f)));
        }
      }
    }
        
    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_3D, texture);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_3D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);

    glTexImage3D(GL_TEXTURE_3D, 0, GL_LUMINANCE,
                 W, W, W, 0,
                 GL_LUMINANCE, GL_UNSIGNED_BYTE, bytes);
  }

  void setup_road(GLuint& texture) {
    int const W = 64, H = 64;
    static unsigned char bytes[W*H*4];

    // TODO: http://www.vterrain.org/Water/textures.html
    for (int i = 0; i < W; ++i) {
      for (int j = 0; j < H; ++j) {
        bool road_edge = i==2 || i==W-3;
        bool center_divider = (i==W/2+1) || (i==W/2-1);
        bool lane = !center_divider
                    && !road_edge
                    && (i != W/2)
                    && i > 2
                    && ((i*8) % W == 0)
                    && j < H/3;
        int alpha = 255;
        if (i < 1 || i > W-2) alpha = 0;

        bytes[i*W*4+j*4] = (road_edge||center_divider||lane)? 255:128;
        bytes[i*W*4+j*4+1] = (road_edge||center_divider||lane)? 255:128;
        bytes[i*W*4+j*4+2] = (road_edge||lane)? 255:128;
        bytes[i*W*4+j*4+3] = alpha;
      }
    }

    glGenTextures(1, &texture);
    glBindTexture(GL_TEXTURE_2D, texture);
    glPixelStorei(GL_UNPACK_ALIGNMENT, 1);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S, GL_REPEAT);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T, GL_CLAMP);
    glTexEnvf(GL_TEXTURE_ENV, GL_TEXTURE_ENV_MODE, GL_MODULATE);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
    glTexParameteri(GL_TEXTURE_2D, GL_GENERATE_MIPMAP_SGIS, GL_TRUE);
    glTexImage2D(GL_TEXTURE_2D, 0, GL_RGBA,
                 W, H, 0,
                 GL_RGBA, GL_UNSIGNED_BYTE, bytes);
  }

  void setup() {
    cout << "Setting up textures." << endl;

    // Idea seen on gamedev.net: 3-dimensional
    // texture where the texture depth goes from rock to water
    // (http://www.gamedev.net/reference/articles/land/landscape_3dtex.jpg).
    // I'm not sure this is any better than multiple layers, except
    // that it automatically deals with transitions.

    setup_road(road);
    
    setup_generic(polished, 32.0f, 0.2f);
    setup_generic(smooth, 16.0f, 0.5f);
    setup_generic(rough, 8.0f, 0.8f);
    setup_volume(contour);
  }

}
