#include "points3.h"
#include "world.h"

namespace Hex {
  float dx[7] =
    {0, 2, 1, -1, -2, -1, 1};
  float dy[7] =
    {0, 0, 1, 1, 0, -1, -1};
}

#if RHOMBUS_GRID
#error Unsupported
#else

barycentric_location::barycentric_location(float x, float y) {
  /* A hexagonal grid can be viewed as a grid of rhombuses, where four
     centers of hexagons form a rhombus.  A rhombus grid can easily be
     transformed into a rectangular grid by shearing.  */
  float ya = y - 0.5*Hex::yHeight*(x/Hex::xWidth);
  /* Now we need to figure out the grid coordinates of the four
     corners of the rhombus.  These two points are always part of the
     triangle, but the third point is determined later.
     */
  m1 = int(floor(x/Hex::xWidth));
  m2 = m1 + 1;
  n1 = 1 + int(floor(ya/Hex::yHeight)) + (m1/2);
  n2 = int(floor(ya/Hex::yHeight)) + (m2/2);

  /* In which half of the rhombus is the point?  This tells us the
     third point of the triangle. */
  float px = frac(x/Hex::xWidth);
  float py = frac(ya/Hex::yHeight);

  if (px + py < 1.0f) {
    m3 = m1; n3 = n1 - 1;
    p1 = py;
    p2 = px;
    p3 = 1.0f - px - py;
  } else {
    m3 = m2; n3 = n2 + 1;
    p1 = 1.0f - px;
    p2 = 1.0f - py;
    p3 = px + py - 1.0f;
  }

  /* Note: barycentric coordinates are computed as follows:

b0 =  (x2 - x1) * (y3 - y1) - (x3 - x1) * (y2 - y1)
b1 = ((x2 - x0) * (y3 - y0) - (x3 - x0) * (y2 - y0)) / b0
b2 = ((x3 - x0) * (y1 - y0) - (x1 - x0) * (y3 - y0)) / b0
b3 = ((x1 - x0) * (y2 - y0) - (x2 - x0) * (y1 - y0)) / b0

  */
}

#endif

float world_to_z(float x, float y) {
  barycentric_location B(x, y);
  float z1 = Map::valid(B.m1, B.n1)? Map::height[B.m1][B.n1] : 0.0f;
  float z2 = Map::valid(B.m2, B.n2)? Map::height[B.m2][B.n2] : 0.0f;
  float z3 = Map::valid(B.m3, B.n3)? Map::height[B.m3][B.n3] : 0.0f;
  float z = z1 * B.p1 + z2 * B.p2 + z3 * B.p3;

  return z;
}

Vector3d world_to_gradient(float x, float y, Map::array field) {
  /* Like world_to_z, but compute the 2d gradient vector instead */

  // TODO: make this work with barycentric_location.
  
  float ya = y - 0.5*Hex::yHeight*(x/Hex::xWidth);
  int mL = int(floor(x/Hex::xWidth));
  int mR = mL + 1;
  int nL = int(floor(ya/Hex::yHeight)) + (mL/2);
  int nR = int(floor(ya/Hex::yHeight)) + (mR/2);

  float z1 = Map::valid(mR, nR)? field[mR][nR] : 0.0f;
  float z2 = Map::valid(mL, nL+1)? field[mL][nL+1] : 0.0f;

  float px = frac(x/Hex::xWidth);
  float py = frac(ya/Hex::yHeight);

  if (px + py < 1.0f) {
    float z0 = Map::valid(mL, nL)? field[mL][nL] : 0.0f;
    return Vector3d(z0 * 0.5f*Hex::yHeight
                    + z1 * -Hex::yHeight
                    + z2 * 0.5f*Hex::yHeight,
                    z0 * Hex::xWidth
                    + z1 * 0.0f
                    + z2 * -Hex::xWidth,
                    0.0f);
  } else {
    float z3 = Map::valid(mR, nR+1)? field[mR][nR+1] : 0.0f;
    return Vector3d(z3 * -0.5f*Hex::yHeight
                    + z1 * -0.5f*Hex::yHeight
                    + z2 * Hex::yHeight,
                    z3 * -Hex::xWidth
                    + z1 * Hex::xWidth
                    + z2 * 0.0f,
                    0.0f);
  }
  /* Note: the gradient vector is the partial derivative of z(x,y)
     with respect to x and y.  I derived these formulas by starting
     with the barycentric representation:

     df/dx = (y2-y3)*z1 + (y3-y1)*z2 + (y1-y2)*z3
     df/dy = (x3-x2)*z1 + (x1-x3)*z2 + (x2-x1)*z3
  */

}

