#ifndef _WORLD_H
#define _WORLD_H

#include <cmath>
#include <vector>
#include "points3.h"

#define RHOMBUS_GRID 0

/* Hexagonal math */
namespace Hex {
  float const t_size = 2.5; // triangle size
  float const xWidth = t_size*sqrt(0.75);
  float const yHeight = t_size;
  
  // The directions for neighboring hexes: (self), SE, NE, N, NW, SW, S

  // These are displacements from hex grid coordinates to neighboring
  // hexes.  Since we are using an array where every other *column* is
  // offset by 0.5*yHeight, we need to know which column is being used
  // to determine what the neighbors are.
#if RHOMBUS_GRID
  static int const dm[7] = // index on [k]
    {0, -1, 0, 1, 1, 0, -1};
  static int const dn[2][7] = { // index on [m&1][k] (note: not n&1)
    {0, 0, 1, 1, 0, -1, -1},
    {0, 0, 1, 1, 0, -1, -1}
  };
#else
  static int const dm[7] = // index on [k]
    {0, 1, 1, 0, -1, -1, 0};
  static int const dn[2][7] = { // index on [m&1][k] (note: not n&1)
    {0, -1, 0, 1, 0, -1, -1},
    {0, 0, 1, 1, 1, 0, -1}
  };
#endif
  
  // These are displacements in pixel coordinates to neighboring hexes.
  extern float neighbor_dx[7], neighbor_dy[7];

  // The directions for vertices: (self), E, NE, NW, W, SW, SE.  Counterclockwise for OpenGL.
  // TODO: should this be 8?  it could save some mod operations later ..

  // Vertex direction k is the place where hex H, H's k-neighbor, and
  // H's (k+1)-neighbor meet.

  // These are displacements from the center to the vertices of a hex,
  // measured in terms of xWidth and yHeight
  extern float dx[7], dy[7];
}

namespace Map {
  int const size = 192;
  typedef float array[size][size];

#ifdef RHOMBUS_GRID
  const float min_x = -(size-1)*Hex::xWidth;
  const float max_x = +(size-1)*Hex::xWidth;
  const float min_y = 0.0f;
  const float max_y = (size-1)*Hex::yHeight;
#else
  const float min_x = 0.0f;
  const float max_x = (size-1)*Hex::xWidth;
  const float min_y = 0.0f;
  const float max_y = (size-1+0.5f)*Hex::yHeight;
#endif
  
  inline bool valid(int m, int n) {
    return m >= 0 && m < size && n >= 0 && n < size;
  }

  /* Simulation state */
  extern array height;

  /* Derived information from the height map */
  extern array nx_base, ny_base, nz_base; // vertex normals
  extern array gx_base, gy_base, gz_base; // vertex gradients
}
  
/* Coordinate conversion from 2d map coordinates to 2d world coordinates and back */
inline Point3d grid_to_world(int m, int n, float z=0.0) {
#if RHOMBUS_GRID
  return Point3d((n-m)*Hex::xWidth,
                 (n+m)*0.5f*Hex::yHeight,
                 z);
#else
  return Point3d(m*Hex::xWidth, (n + (m&1)*0.5f)*Hex::yHeight, z);
#endif
}

inline std::pair<int,int> world_to_grid(float x, float y) {
#if RHOMBUS_GRID
  int diff = int(round(x/Hex::xWidth));
  int sum = int(round(y/(0.5f*Hex::yHeight)));
  int m = (sum - diff)/2;
  int n = (sum + diff)/2;
#else
  int m = int(round(x/Hex::xWidth));
  int n = int(round(y/Hex::yHeight - (m&1)*0.5f));
#endif
  return std::make_pair(m, n);
}

inline float frac(float x) {
  return x - floor(x);
}

// The barycentric location tells us where inside a triangle a
// position lies.  Given x and y, we can compute the grid triangle
// containing the point as well as the barycentric coordinates of the
// point within the triangle.  p1 + p2 + p3 == 1.0, and the m,n values
// are all part of one triangle.
struct barycentric_location {
  barycentric_location(float x, float y);

  int m1, n1; float p1;
  int m2, n2; float p2;
  int m3, n3; float p3;
};

/* Compute the z coordinate on the landscape corresponding to an (x,y)
   position.  This finds the triangle inside which (x,y) is located,
   and computes z from there. */
float world_to_z(float x, float y);

/* Compute the gradient vector, which tells us something about the
   slope at (x,y). */
Vector3d world_to_gradient(float x, float y, Map::array field);

#endif
